/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Adblock Plus.
 *
 * The Initial Developer of the Original Code is
 * Wladimir Palant.
 * Portions created by the Initial Developer are Copyright (C) 2006-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * ***** END LICENSE BLOCK ***** */

var EXPORTED_SYMBOLS = [];

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");

// Gecko 1.9.0/1.9.1 compatibility - add XPCOMUtils.defineLazyServiceGetter
if (!("defineLazyServiceGetter" in XPCOMUtils))
{
	XPCOMUtils.defineLazyServiceGetter = function XPCU_defineLazyServiceGetter(obj, prop, contract, iface)
	{
		obj.__defineGetter__(prop, function XPCU_serviceGetter()
		{
			delete obj[prop];
			return obj[prop] = Cc[contract].getService(Ci[iface]);
		});
	};
}

// Hack Utils.appID and Utils.platformVersion, K-Meleon doesn't have XULAppInfo
Cu.import("resource:///modules/adblockplus/Utils.jsm");
Utils.__defineGetter__("appID", function() "kmeleon@sf.net");
Utils.__defineGetter__("platformVersion", function() "1.9.1");

// Hack Utils.windowMediator, we want our own window enumeration
var mediator =
{
	__proto__: Utils.windowMediator,
	getZOrderDOMWindowEnumerator: function()
	{
		return {
			enumerator: windows.allWindows(),
			nextValue: undefined,
			hasMoreElements: function()
			{
				try
				{
					if (typeof this.nextValue == "undefined")
						this.nextValue = this.enumerator.next();
					return true;
				}
				catch (e if e instanceof StopIteration)
				{
					return false;
				}
			},
			getNext: function()
			{
				if (typeof this.nextValue == "undefined")
					this.nextValue = this.enumerator.next();
				let result = this.nextValue;
				this.nextValue = undefined;
				return result;
			}
		}
	}
}
Utils.__defineGetter__("windowMediator", function() mediator);

// Start up Bootstrap.jsm, it will initialize everything else
{
	Cu.import("resource:///modules/adblockplus/Bootstrap.jsm");
	Bootstrap.startup();
}

// Now we can load all the other modules we need
Cu.import("resource:///modules/adblockplus/AppIntegration.jsm");

// Detect ANSI codepage to be used - yes, K-Meleon still doesn't support Unicode
let unicodeConverter = Cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Ci.nsIScriptableUnicodeConverter);
unicodeConverter.charset = "windows-1252";
try
{
	let xulRegistry = Cc["@mozilla.org/chrome/chrome-registry;1"].getService(Ci.nsIXULChromeRegistry);
	let locale = xulRegistry.getSelectedLocale("adblockplus");
	let charsets = {
		ar: "windows-1256",
		bg: "windows-1251",
		cs: "windows-1250",
		el: "windows-1253",
		et: "windows-1257",
		he: "windows-1255",
		hr: "windows-1250",
		hu: "windows-1250",
		ja: "Shift_JIS",
		'kk-KZ': "windows-1251",
		ko: "EUC-KR",
		lt: "windows-1257",
		lv: "windows-1257",
		mn: "windows-1251",
		pl: "windows-1250",
		sk: "windows-1250",
		sl: "windows-1250",
		sr: "windows-1251",
		th: "windows-874",
		tr: "windows-1254",
		ru: "windows-1251",
		uk: "windows-1251",
		vi: "windows-1258",
		'zh-CN': "windows-936",
		'zh-TW': "Big5"
	};
	if (locale in charsets)
		unicodeConverter.charset = charsets[locale];
}
catch(e)
{
	Cu.reportError(e);
}

// Replace ObjectTabs.getHooksForElement() by our own function
Cu.import("resource:///modules/adblockplus/ObjectTabs.jsm", null).objTabs.getHooksForElement = function(element)
{
	let tab = Utils.getWindow(element);
	let wnd = windows.getWindowForTab(tab);
	if (wnd)
		return wnd.wrapper.E("abp-hooks");
	else
		return null;
}

function convertUIString(str)
{
	str = str.replace(/\u2026/g, '...');
	return unicodeConverter.ConvertFromUnicode(str);
}

// Register quit observer to shut down Bootstrap.jsm when necessary
{
	let observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
	let observer = {
		observe: function(subject, topic, data)
		{
			if (topic == "quit-application")
			{
				observerService.removeObserver(this, "quit-application");
				Bootstrap.shutdown(false);
			}
		},

		QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference]),
	};
	observerService.addObserver(observer, "quit-application", true);
}

function fakeAddEventListener(eventName, handler, capture)
{
	if (!("_handlers_" + eventName in this))
		this["_handlers_" + eventName] = [];
	handlers = this["_handlers_" + eventName];
	if (handlers.indexOf(handler) < 0)
		handlers.push(handler);
}
function fakeRemoveEventListener(eventName, handler, capture)
{
	if (!("_handlers_" + eventName in this))
		return;
	handlers = this["_handlers_" + eventName];
	let index = handlers.indexOf(handler);
	if (index >= 0)
		handlers.splice(index, 1)
}
function fakeDispatchEvent(event)
{
	if (!("_handlers_" + event.type in this))
		return;
	handlers = this["_handlers_" + event.type];
	for (let i = 0; i < handlers.length; i++)
		handlers[i].call(this, event);
}

/**
 * Provide a valid nsIXULAppInfo object for the issue reporter.
 */
delete Utils.appInfo;
Utils.appInfo =
{
	vendor: "K-Meleon",
	name: "K-Meleon",
	ID: "",
	get version() getKMeleonVersion(),
	appBuildID: "",
	platformVersion: "",
	platformBuildID: ""
};

/**
 * Load fake window document and apply the usual overlay to it.
 */
var fakeDoc = null;
{
	let request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
	request.open("GET", "chrome://adblockplus/content/ui/kmeleonFakeWindow.xul", false);
	request.send(null);

	fakeDoc = Cc["@mozilla.org/xul/xul-document;1"].createInstance(Ci.nsIDOMXULDocument);
	fakeDoc.appendChild(request.responseXML.documentElement);

	request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
	request.open("GET", "chrome://adblockplus/content/ui/overlayGeneral.xul", false);
	request.send(null);

	// Apply overlay "manually", document.loadOverlay() crashes
	for (let node = request.responseXML.documentElement.firstChild; node; node = node.nextSibling)
	{
		if (node.nodeType != Ci.nsIDOMNode.ELEMENT_NODE || !node.hasAttribute("id"))
			continue;

		let insertionPoint = fakeDoc.getElementById(node.getAttribute("id"));
		if (!insertionPoint)
			continue;

		for (let i = 0; i < node.attributes.length; i++)
		{
			attribute = node.attributes[i];
			insertionPoint.setAttribute(attribute.nodeName, attribute.nodeValue)
		}
		while (node.firstChild)
			insertionPoint.appendChild(node.firstChild);
	}

	// We are already initialized, no further initialization necessary
	fakeDoc.getElementById("abp-hooks").setAttribute("currentVersion", Utils.addonVersion);
}

/**
 * Object tracking windows and tabs.
 */
var windows =
{
	windows: {},

	init: function()
	{
		Utils.windowWatcher.registerNotification(this);
	},

	addWindow: function(hWnd)
	{
		if (hWnd in this.windows)
			return;

		let wnd = new FakeWindow(hWnd);
		AppIntegration.addWindow(wnd);
		wnd.wrapper = AppIntegration.getWrapperForWindow(wnd);
		this.windows[hWnd] = wnd;
	},

	removeWindow: function(hWnd)
	{
		if (!(hWnd in this.windows))
			return;

		this.windows[hWnd].dispatchEvent({type: "unload"});
		delete this.windows[hWnd];
	},

	observe: function(subject, topic, data)
	{
		// Only look at content windows (tabs), not chrome windows
		if (!(subject instanceof Ci.nsIDOMWindow) && !(subject instanceof Ci.nsIDOMChromeWindow))
			return;

		if (topic == "domwindowopened")
		{
			if (subject instanceof Ci.nsIDOMChromeWindow)
				addRootListener(subject, "load", true);
			else
			{
				addRootListener(subject, "focus", true);
				addRootListener(subject, "contextmenu", true);
				addRootListener(subject, "click", true);
			}
		}
	},

	allWindows: function()
	{
		for each (wnd in this.windows)
			yield wnd;
	},

	getWindow: function(hWnd)
	{
		if (hWnd in this.windows)
			return this.windows[hWnd];
		else
			return null;
	},

	getWindowForTab: function(tab)
	{
		try
		{
			let hWnd = getHWND(Utils.windowWatcher.getChromeForWindow(tab).QueryInterface(Ci.nsIEmbeddingSiteWindow));
			return this.getWindow(hWnd);
		}
		catch (e)
		{
			return null;
		}
	},

	QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference])
}

// Fake window/document/browser to be given to the AppIntegration module
function FakeWindow(hWnd)
{
	this.hWnd = hWnd;
	this.iconContextMenuItems = {__proto__: null};

	// Create a copy of the fake document, icon attributes are per-window
	this.document = Cc["@mozilla.org/xul/xul-document;1"].createInstance(Ci.nsIDOMXULDocument);
	this.document.appendChild(this.document.importNode(fakeDoc.documentElement, true));

	let me = this;
	this.document.__defineGetter__("defaultView", function() me);

	// Display status bar icon
	if (!this.document.getElementById("abp-status").hidden)
		setStatusbarIcon(this.hWnd, this.icon);

	// Get notified if the icon is updated or status bar icon is hidden
	let me = this;
	this.document.getElementById("abp-toolbarbutton").addEventListener("DOMAttrModified", function()
	{
		me.updateIconDelayed();
	}, false);
	this.document.getElementById("abp-status").addEventListener("DOMAttrModified", function(event)
	{
		if (event.attrName == "hidden")
		{
			if (event.newValue == "true")
				hideStatusbarIcon(hWnd);
			else
				setStatusbarIcon(hWnd, me.icon);
		}
	}, false);
}
FakeWindow.prototype =
{
	QueryInterface: function() this,
	getInterface: function() this,

	hWnd: null,
	hToolbar: null,
	wrapper: null,
	currentTab: null,
	iconContextMenuItems: null,
	icon: 0,
	scheduledIcon: 0,

	allowSubframes: true,

	addEventListener: fakeAddEventListener,
	removeEventListener: fakeRemoveEventListener,
	dispatchEvent: fakeDispatchEvent,

	focus: function()
	{
		focusWindow(this.hWnd);
	},

	triggerOverlayEvent: function(element, event)
	{
		if (typeof element == "string")
		{
			element = this.document.getElementById(element);
			if (!element)
				return;
		}

		if (typeof event == "string")
		{
			let eventType = event;
			event = this.document.createEvent("UIEvents");
			event.initUIEvent(eventType, false, true, null, 0);
		}
		element.dispatchEvent(event);
	},

	addMenuItems: function(popup)
	{
		let menu = createPopupMenu();

		for (let child = popup.firstChild; child; child = child.nextSibling)
		{
			if (child.nodeType != Ci.nsIDOMNode.ELEMENT_NODE || child.hidden)
				continue;

			// We should not show "show in toolbar" option
			if (child.tagName == "menuitem" && /-showintoolbar$/.test(child.id))
				continue;

			let type = 0;
			if (child.tagName == "menuseparator")
				type = -1;
			else if (child.tagName == "menu")
				type = this.addMenuItems(child.getElementsByTagName("menupopup")[0]);

			if (!("menuID" in child))
			{
				if (child.tagName == "menuitem")
				{
					child.menuID = createCommandID();
					this.iconContextMenuItems[child.menuID] = child;
				}
				else
					child.menuID = -1;
			}

			addMenuItem(menu, type, child.menuID,
											convertUIString(child.getAttribute("label")),
											child.getAttribute("default") == "true",
											child.getAttribute("disabled") == "true",
											child.getAttribute("checked") == "true");

			// Toggle checkbox selection so if it is clicked we get the right value
			if (child.getAttribute("type") == "checkbox")
			{
				if (child.getAttribute("checked") == "true")
					child.removeAttribute("checked");
				else
					child.setAttribute("checked", "true");
			}
		}
		return menu;
	},

	triggerMenuItem: function(id)
	{
		if (!(id in this.iconContextMenuItems))
			return;

		let menuItem = this.iconContextMenuItems[id];
		if (menuItem.hasAttribute("command"))
			this.triggerOverlayEvent(menuItem.getAttribute("command"), "command");
		else
			this.triggerOverlayEvent(menuItem, "command");
	},

	addProgressListener: function(listener)
	{
		this.addEventListener("locationChange", listener.onLocationChange, false);
	},
	removeProgressListener: function(listener)
	{
		this.removeEventListener("locationChange", listener.onLocationChange, false);
	},

	get contentWindow() this.currentTab || {location: {href: "about:blank"}},
	get content() this.contentWindow,
	get location() this.contentWindow.location,

	get currentURI() Utils.makeURI(this.contentWindow.location.href),

	openDialog: function(url, windowName, features)
	{
		let params = null;
		if (arguments.length > 3)
		{
			params = Cc["@mozilla.org/array;1"].createInstance(Components.interfaces.nsIMutableArray);
			for (let i = 3; i < arguments.length; i++)
			{
				let variant = Cc["@mozilla.org/variant;1"].createInstance(Ci.nsIWritableVariant);
				variant.setAsInterface(Ci.nsIVariant, arguments[i]);
				params.appendElement(variant, false);
			}
		}

		let result = Utils.windowWatcher.openWindow(null, url, windowName, features, params);
		let mainWin = this;
		result.__defineGetter__("opener", function() mainWin);
		return result;
	},

	setCurrentTab: function(tab)
	{
		if (tab == this.currentTab)
			return;

		this.currentTab = tab;
		this.dispatchEvent({type: "select"});
		this.dispatchEvent({type: "locationChange"});
	},

	getBrowser: function()
	{
		return this;
	},

	addTab: function(url)
	{
		openTab(url, this.hWnd);
	},

	invalidateIcon: function()
	{
		if (this.hToolbar)
			invalidateToolbar(this.hToolbar);
		if (!this.document.getElementById("abp-status").hidden)
		{
			hideStatusbarIcon(this.hWnd);
			setStatusbarIcon(this.hWnd, this.icon);
		}
	},

	updateIconDelayed: function()
	{
		// Icon needs to be updated delayed to avoid flickering because we might
		// have two subsequent attribute changes
		let state = this.document.getElementById("abp-toolbarbutton").getAttribute("abpstate");
		this.scheduledIcon = 0;
		if (state == "disabled" || state == "whitelisted")
			this.scheduledIcon = 1;
		Utils.runAsync(function()
		{
			if (this.scheduledIcon == this.icon)
				return;

			this.icon = this.scheduledIcon;
			if (this.hToolbar)
				setToolbarIcon(this.hToolbar, this.icon);
			if (!this.document.getElementById("abp-status").hidden)
				setStatusbarIcon(this.hWnd, this.icon);
		}, this);
	}
};

// Initialization
windows.init();

// Entry points that the DLL will call
function onCommand(command, hWnd, id)
{
	let wnd = null;
	if (typeof hWnd != "undefined")
	{
		wnd = windows.getWindow(hWnd);
		if (!wnd)
			return;
	}

	switch (command)
	{
		case "blockable":
			wnd.wrapper.executeAction(1);
			break;
		case "settings":
			wnd.wrapper.executeAction(2);
			break;
		case "enable":
			wnd.wrapper.executeAction(3);
			break;
		case "sendreport":
			wnd.wrapper.openReportDialog();
			break;
		case "setToolbar":
			wnd.hToolbar = id;
			break;
		case "invalidateIcon":
			for (let wnd in windows.allWindows())
				wnd.invalidateIcon();
			break;
		case "removeWhitelist":
			wnd.triggerOverlayEvent("abp-removeWhitelist-menuitem", "command");
			break;
		case "frame":
			wnd.triggerOverlayEvent("abp-frame-menuitem", "command");
			break;
		case "object":
			wnd.triggerOverlayEvent("abp-object-menuitem", "command");
			break;
		case "media":
			wnd.triggerOverlayEvent("abp-media-menuitem", "command");
			break;
		case "image":
			wnd.triggerOverlayEvent("abp-image-menuitem", "command");
			break;
		case "toolbar":
			if (wnd.document.getElementById("abp-toolbarbutton").getAttribute("type") == "menu")
				showIconContext(wnd.hWnd, false);
			else
				wnd.triggerOverlayEvent("abp-toolbarbutton", "command");
			break;
		case "statusbar":
			if (wnd.document.getElementById("abp-status").hasAttribute("popup"))
				showIconContext(wnd.hWnd, true);
			else
			{
				let event = wnd.document.createEvent("MouseEvents");
				event.initMouseEvent("click", false, true, null, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
				wnd.triggerOverlayEvent("abp-status", event);
			}
			break;
		case "menu":
			wnd.triggerMenuItem(id);
			break;
	}
}

function onBrowserWindowOpened(hWnd)
{
	windows.addWindow(hWnd);
}

function onBrowserWindowClosed(hWnd)
{
	windows.removeWindow(hWnd);
}

function onEvent(event)
{
	if (event.type == "contextmenu")
	{
		let tab = event.target.ownerDocument.defaultView.QueryInterface(Ci.nsIInterfaceRequestor)
										 .getInterface(Ci.nsIWebNavigation)
										 .QueryInterface(Ci.nsIDocShellTreeItem)
										 .rootTreeItem
										 .QueryInterface(Ci.nsIInterfaceRequestor)
										 .getInterface(Ci.nsIDOMWindow);
		let wnd = windows.getWindowForTab(tab);
		if (wnd)
		{
			resetContextMenu();

			wnd.document.__defineGetter__("popupNode", function() event.target);
			wnd.wrapper.updateContextMenu({__proto__: event, eventPhase: Ci.nsIDOMEvent.AT_TARGET});
			delete wnd.document.popupNode;

			let contextMenu = wnd.document.getElementById("contentAreaContextMenu")
			for (let element = contextMenu.firstChild, elementIndex = 0; element; element = element.nextSibling)
			{
				if (element.nodeType != Ci.nsIDOMNode.ELEMENT_NODE)
					continue;
				if (!element.hidden)
					addContextMenuItem(elementIndex, convertUIString(element.getAttribute("label")));
				elementIndex++;
			}
		}
	}
	else if (event.type == "focus" && event.target instanceof Components.interfaces.nsIDOMDocument)
	{
		let tab = event.target.defaultView;
		let wnd = windows.getWindowForTab(tab);
		if (wnd)
			wnd.setCurrentTab(tab);
	}
	else if (event.type == "click")
	{
		let tab = Utils.getWindow(event.target);
		let wnd = windows.getWindowForTab(tab);
		if (wnd)
			wnd.dispatchEvent(event);
	}
	else if (event.type == "load")
	{
		let wnd = event.target.defaultView.top;
		removeRootListener(wnd, "load", true);
		if (wnd.location.href.substr(0, 21) == "chrome://adblockplus/")
		{
			// Make sure that window.close will fire "unload" event
			// http://kmeleon.sourceforge.net/bugs/viewbug.php?bugid=1297
			wnd._oldCloseFunc = wnd.close;
			wnd.close = function()
			{
				let event = wnd.document.createEvent("Events");
				event.initEvent("unload", false, false);
				wnd.dispatchEvent(event);
				wnd._oldCloseFunc();
			}

			// Make sidebar window the top-most window
			if (wnd.location.href.indexOf("sidebarDetached.xul") >= 0)
				setTopmostWindow(getHWND(Utils.windowWatcher.getChromeForWindow(wnd).QueryInterface(Ci.nsIEmbeddingSiteWindow)));

			// Restore persisted location and size. Have to go through nsIBaseWindow
			// because usual moveTo()/resizeTo() will overwrite the results of each
			// other in K-Meleon.
			Utils.runAsync(function()
			{
				wnd.sizeToContent();
				let docElement = wnd.document.documentElement;
				let width = parseInt(docElement.getAttribute("width"), 10) || wnd.outerWidth;
				let height = parseInt(docElement.getAttribute("height"), 10) || wnd.outerHeight;
				let screenX = parseInt(docElement.getAttribute("screenX"), 10) || (wnd.screen.availLeft + (wnd.screen.availWidth - width) / 2);
				let screenY = parseInt(docElement.getAttribute("screenY"), 10) || (wnd.screen.availTop + (wnd.screen.availHeight - height) / 2);
				let baseWindow = wnd.QueryInterface(Ci.nsIInterfaceRequestor)
														.getInterface(Ci.nsIWebNavigation)
														.QueryInterface(Ci.nsIDocShellTreeItem)
														.treeOwner
														.QueryInterface(Ci.nsIBaseWindow);
				baseWindow.setPositionAndSize(screenX, screenY, width, height, false);

				// Check for updates in position and size regularly
				if (docElement.hasAttribute("persist"))
				{
					wnd.setInterval(function()
					{
						if (wnd.outerWidth != width)
						{
							width = wnd.outerWidth;
							docElement.setAttribute("width", width);
							wnd.document.persist(docElement.id, "width");
						}
						if (wnd.outerHeight != height)
						{
							height = wnd.outerHeight;
							docElement.setAttribute("height", height);
							wnd.document.persist(docElement.id, "height");
						}
						if (wnd.screenX != screenX)
						{
							screenX = wnd.screenX;
							docElement.setAttribute("screenX", screenX);
							wnd.document.persist(docElement.id, "screenX");
						}
						if (wnd.screenY != screenY)
						{
							screenY = wnd.screenY;
							docElement.setAttribute("screenY", screenY);
							wnd.document.persist(docElement.id, "screenY");
						}
					}, 1000);
				}
			});
		}
	}
}

function getTooltipText(hWnd, status, unicode)
{
	let wnd = windows.getWindow(hWnd);
	if (!wnd)
		return null;

	wnd.document.__defineGetter__("tooltipNode", function() wnd.document.getElementById(status ? "abp-status" : "abp-toolbarbutton"));
	wnd.wrapper.fillTooltip({});
	delete wnd.document.tooltipNode;

	let tooltipText = "";
	for (let node = wnd.document.getElementById("abp-tooltip").firstChild; node; node = node.nextSibling)
	{
		if (node.nodeType != Ci.nsIDOMNode.ELEMENT_NODE || node.hidden)
			continue;

		if (node.hasAttribute("value"))
			tooltipText += node.getAttribute("value") + (node.localName == "description" ? "\n\n" : "\n");

		for (let child = node.firstChild; child; child = child.nextSibling)
			if (child.hasAttribute("value"))
				tooltipText += child.getAttribute("value") + "\n";
	}
	return tooltipText;
}

function buildContextMenu(hWnd, status)
{
	let wnd = windows.getWindow(hWnd);
	if (!wnd)
		return null;

	wnd.document.__defineGetter__("popupNode", function() wnd.document.getElementById(status ? "abp-status" : "abp-toolbarbutton"));
	wnd.wrapper.fillPopup({target: wnd.document.getElementById("abp-toolbar-popup")});
	delete wnd.document.popupNode;

	return wnd.addMenuItems(wnd.document.getElementById("abp-toolbar-popup"));
}
